/*****************************************************************************
* Author: Paulino Ruiz de Clavijo Vázquez <paulino@dte.us.es>
* Date: 05-12-2012
* Revision: 0.3
* Project: Nanofs
/** -- End auto header, don't touch this line -- ****************************/
 
 


#include <linux/fs.h>

#include <linux/buffer_head.h>
#include "../include/nanofs.h"

void dump_bh(struct buffer_head *bh,int max_dump);

/** Read dir_entry from block device
 *  node_no is offset in blocks, dir_node begins at start of block
 *  Cannot read struct from device directly, struct is is aligned is memory but not in device
 */
int nanofs_fill_dir_node(struct super_block *sb,unsigned long node_no, struct nanofs_dir_node *dn)
{
    struct buffer_head *bh;

    sb_min_blocksize(sb, NANOFS_BLKSIZE); // Try set bs=512
    bh = sb_bread(sb, node_no);
    if (!bh)
    {
        printk(KERN_DEBUG "nanofs-fill-dir-node: Error reading block device at node %ld",node_no);
        return -EIO;
    }
    dn->d_flags=(__u8)*(bh->b_data);
    dn->d_next_ptr=le32_to_cpu(* (__u32 *)(bh->b_data + 1));
    dn->d_data_ptr=le32_to_cpu(* (__u32 *)(bh->b_data + 5));
    dn->d_meta_ptr=le32_to_cpu(* (__u32 *)(bh->b_data + 9));

    dn->d_fname_len = (__u8)*(bh->b_data + 13);
    if (dn->d_fname_len>NANOFS_MAXFILENAME)
    {
        brelse(bh);
        printk (KERN_ERR "NANOFS: Not valid filename length on dir_node at block %lxH.\n",
                node_no);
        return -1;
    }
    // Copy filename
    memcpy(dn->d_fname,bh->b_data + 14,dn->d_fname_len);

    printk(KERN_DEBUG "nanofs-fill-dir-node: dir_node at %lxH, f=%xH,next=%xH,data=%xH,meta=%xH,len=%d",
           node_no,dn->d_flags,dn->d_next_ptr,dn->d_data_ptr,dn->d_meta_ptr,dn->d_fname_len);

    if (dn->d_fname_len<NANOFS_MAXFILENAME) 
        dn->d_fname[dn->d_fname_len]=0;
    printk (KERN_ERR "nanofs-fill-dir-node: fname '%s'\n",dn->d_fname);


    //dump_bh(bh,20);
    brelse(bh);
    return 0;
}

/** Return file size */
long nanofs_get_file_size(struct super_block *sb,struct nanofs_dir_node *dn)
{
    long res=0;
    //unsigned long data_blkno=0;
    struct nanofs_data_node data_node;
    struct buffer_head *bh;

    if (DN_ISDIR((*dn)))
        return NANOFS_BLKSIZE;
    if (dn->d_data_ptr==0)
        return 0;
    sb_min_blocksize(sb, NANOFS_BLKSIZE); // Try set bs=512
    // Count data blocks
    data_node.d_next_ptr=dn->d_data_ptr;
    while (data_node.d_next_ptr)
    {
        bh = sb_bread(sb, data_node.d_next_ptr);
        if (!bh)
        {
            printk(KERN_ERR "nanofs-get-file-size: Error reading block device at blk_no %xlH\n",data_node.d_next_ptr);
            return -EIO;
        }
        printk(KERN_DEBUG "nanofs-get-file-size: Data blk_no %lxH\n",data_node.d_next_ptr);
        //dump_bh(bh,20);
        data_node.d_next_ptr=le32_to_cpu(*(__u32 *)(bh->b_data));
        data_node.d_len=le32_to_cpu(*(__u32 *)(bh->b_data+4));
        res+=data_node.d_len;
        brelse(bh);
        printk(KERN_DEBUG "nanofs-get-file-size: next_ptr %lxH, size %ld\n",data_node.d_next_ptr,data_node.d_len);

    }
    return res;

}

/** Debug functions */

void dump_bh(struct buffer_head *bh,int max_dump)
{
    int i;
    if (bh)
    {
        printk(KERN_DEBUG "nanofs-dumpbh: b_size=%ld",bh->b_size);
        for (i=0;i<bh->b_size-1 && i<max_dump;i++)
            printk(KERN_DEBUG "nanofs-dumpbh: %3.3x=%2.2x",i,(unsigned char)bh->b_data[i]);

    }

}
