/*****************************************************************************
* Author: Paulino Ruiz de Clavijo Vázquez <paulino@dte.us.es>
* Date: 05-12-2012
* Revision: 0.3
* Project: Nanofs
/** -- End auto header, don't touch this line -- ****************************/
 
 


/*
 * Copyright (c) 2011 Paulino Ruiz de Clavijo Vázquez <paulino@dte.us.es>
 *
 * Experimental FS
 *
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>

#include <linux/kernel.h>
#include <linux/blkdev.h>

#include "../include/nanofs.h"

/// @TODO: Only implemented nanofs of 512bytes blksize
#define NANOFS_BLKSIZE 512
//#define NANOFS_REVISION 1


#include "blk.c"

MODULE_AUTHOR ( "Paulino Ruiz de Clavijo" );
MODULE_DESCRIPTION ( "Nanofs" );
MODULE_LICENSE ( "GPL" );



static struct inode *nanofs_get_inode ( struct super_block *sb, unsigned long pos );
static int nanofs_readdir ( struct file *filp, void *dirent, filldir_t filldir );
static struct dentry *nanofs_lookup ( struct inode *parent_inode, struct dentry *dentry, struct nameidata *nd );
static int nanofs_readpage ( struct file *file, struct page *page );
static int nanofs_writepage(struct page *page, struct writeback_control *wbc);



/* Superblock */

//static void nanofs_put_super ( struct super_block *sb );
static void nanofs_kill_sb ( struct super_block *sb );

static struct super_operations nanofs_ops = {
    //.alloc_inode  = nanofs_alloc_inode,
    //.destroy_inode    = nanofs_destroy_inode,
    .statfs         = simple_statfs,
    .drop_inode     = generic_delete_inode
    //.put_super      = nanofs_put_super
};

/* File operations */

static struct file_operations nanofs_file_ops = {
    .open   = generic_file_open,
    //.read = do_sync_read, // Fails, it does "rw_verify_area" ...
    .aio_read       = generic_file_aio_read,
    //.splice_read      = generic_file_splice_read,
    .llseek = generic_file_llseek,
    .mmap   = generic_file_mmap,
    //fsync:
    //.write:generic_file_write,
};

static const struct file_operations nanofs_dir_ops = {
    .read       = generic_read_dir,
    .readdir    = nanofs_readdir,
    .llseek     = default_llseek,
};


/* Inode operations */

static const struct inode_operations nanofs_dir_inode_ops = {
    .lookup     = nanofs_lookup,
};

/* Read cache pages */
static const struct address_space_operations nanofs_aops = {
    .readpage   = nanofs_readpage,
    .writepage  = nanofs_writepage
};


/*
 * allocate a new inode
 */
/*static struct inode *nanofs_alloc_inode(struct super_block *sb)
{
    //struct nanofs_inode_info *inode;
    void *inode;
    inode = kmem_cache_alloc(512, GFP_KERNEL);
    return inode ? &inode->vfs_inode : NULL;
}*/
static int nanofs_writepage(struct page *page, struct writeback_control *wbc)
{
    return 0;
    //return block_write_full_page(page, fat_get_block, wbc);
}



static int nanofs_readpage ( struct file *file, struct page *page )
{
    struct buffer_head *bh;
    struct inode *inode = page->mapping->host;
    struct super_block *sb=inode->i_sb;
    loff_t offset, size;
    long fillsize;//, pos;
    void *buf;
    int ret=0;
    struct nanofs_dir_node dn;
    struct nanofs_data_node data_node;
    
    long int blk_no;
    int bytes_read;


    buf = kmap ( page );
    if ( !buf )
        return -ENOMEM;
    
    sb_min_blocksize(sb, NANOFS_BLKSIZE); // Try set bs=512bytes
    
    // Reading first data_node of file
    ret=nanofs_fill_dir_node(sb,inode->i_ino,&dn);
    blk_no=dn.d_data_ptr;
    bh = sb_bread(sb, blk_no);
    if(!bh)
        return -EIO;
    data_node.d_next_ptr=le32_to_cpu(*(__u32 *)(bh->b_data));
    data_node.d_len=le32_to_cpu(*(__u32 *)(bh->b_data+4));
    printk (KERN_DEBUG "nanofs-readpage: data_node next_ptr=%lxH,len=%ld\n",
              data_node.d_next_ptr,data_node.d_len);
    brelse(bh);

    /* 32 bit warning */
    offset = page_offset ( page );
    size = i_size_read ( inode );
    fillsize = 0;
    ret = 0;
    printk (KERN_DEBUG "nanofs-readpage: page_index=%d,offset=%lld,inode_no=%lxH,size=%lld\n",
             ( int ) page->index,offset,inode->i_ino,size );

    // Getting data
    if (offset < size)
    {
        size -= offset;
        fillsize = size > PAGE_SIZE ? PAGE_SIZE : size;
        //pos=0;
        // next node in data list
        
        if(data_node.d_len < offset + fillsize)
        {
            //TODO:Navigate over link list
            printk (KERN_DEBUG "nanofs-readpage: file fragmentation not implemented\n");
            ret= -EIO;
        }
        else
        {
            blk_no+= (offset + 8) >> (sb->s_blocksize_bits);
            memset(buf, 0, PAGE_SIZE);  // Really is this necesary?
            bytes_read=0;
            // Cache PAGE_SIZE=4096 and NANOFS_READSIZE = 512
            while(fillsize > 0)
            {
                printk (KERN_DEBUG "nanofs-readpage: reading blk_no=%xd, fillsize=%d\n",
                    blk_no,fillsize);
                bh = sb_bread(inode->i_sb, blk_no);
                if(!bh)
                {
                    SetPageError(page);
                    ret=-EIO;
                    break;
                }
                if(bytes_read == 0) // first block
                {
                    memcpy(buf + bytes_read, bh->b_data + 8, 
                           fillsize > NANOFS_BLKSIZE -8 ? NANOFS_BLKSIZE -8 : fillsize);
                    bytes_read=fillsize > NANOFS_BLKSIZE -8 ? NANOFS_BLKSIZE -8 : fillsize;
                    fillsize-=bytes_read;
                }
                else // Other blocks
                {
                    memcpy(buf + bytes_read, bh->b_data , fillsize > NANOFS_BLKSIZE  ? NANOFS_BLKSIZE : fillsize);
                    fillsize-=NANOFS_BLKSIZE;
                    bytes_read+=NANOFS_BLKSIZE;
                }
                blk_no++;
                brelse(bh);
            }
        }                
    }    
 
    if ( ret == 0 )
        SetPageUptodate ( page );

    flush_dcache_page ( page );
    kunmap ( page );
    if ( PageLocked ( page ) )
        unlock_page ( page );
    return ret;
}

/*
 * look up an entry in a directory
 */
static struct dentry *nanofs_lookup ( struct inode *dir_inode, 
                      struct dentry *dentry, struct nameidata *nd ) {
    
    struct inode *new_inode;
    struct nanofs_dir_node dn;
    unsigned long current_ino;

    //printk(KERN_DEBUG "nanofs: directory look_up\n");
    /*if(dir_inode->i_ino != 1) // Only from root node "1" is root node
        return ERR_PTR(-ENOENT);*/
    
    printk ( KERN_DEBUG "nanofs-lookup: i_ino %lxH, fname='%s'\n",
             dir_inode->i_ino,dentry->d_name.name );
    if ( nanofs_fill_dir_node ( dir_inode->i_sb,dir_inode->i_ino,&dn ) ) // Read parent dir
    {
        printk ( KERN_DEBUG "nanofs-lookup: Error getting parent node at i_ino %lxH\n",dir_inode->i_ino );
        return ERR_PTR(-1);
    }
    current_ino=dn.d_data_ptr;
    while ( current_ino ) {
        if ( nanofs_fill_dir_node ( dir_inode->i_sb,current_ino,&dn ) ) // Read parent dir
        {
            printk ( KERN_DEBUG "nanofs-lookup: Error navigating dir at ino %lxH\n",current_ino);
            return ERR_PTR(-1);
        }
        //printk ( KERN_DEBUG "nanofs-lookup: cmp \n",current_ino );
        if ( strncmp ( dentry->d_name.name,dn.d_fname,dn.d_fname_len ) == 0 )
            break;
        current_ino=dn.d_next_ptr;
    }

    if ( current_ino ) { // Found
        new_inode=nanofs_get_inode ( dir_inode->i_sb,current_ino );
        if ( !new_inode )
            return ERR_PTR ( -EACCES );
        printk ( KERN_DEBUG "nanofs-lookup: dir entry found at ino %ld\n",current_ino );
        d_add ( dentry, new_inode );
        return ERR_PTR(0);
    }
    /*
     * it's a bit funky, _lookup needs to return an error code
     * (negative) or a NULL, both as a dentry.  ENOENT should not
     * be returned, instead we need to create a negative dentry by
     * d_add(dentry, NULL); and return 0 as no error.
     * (Although as I see, it only matters on writable file
     * systems).
     */
    printk ( KERN_DEBUG "nanofs-lookup: file not found at ino %ld\n",current_ino );
    d_add ( dentry, NULL );
    return ERR_PTR(0);
}

/*
 * read the entries from a directory
 */
static int nanofs_readdir ( struct file *filp, void *dirent, filldir_t filldir )
{
    struct inode *i = filp->f_dentry->d_inode;
    //struct dentry *de = filp->f_dentry;
    struct nanofs_dir_node dn;
    //unsigned long offset=filp->f_pos;;
    ino_t ino;
    //int res;


    //offset=filp->f_pos;
    //if(!offset) // First directory entry
    //{
    //offset= i->i_ino;
    //}
    //else
    //  return 1;
    if ( filp->f_pos>0 )
        return 1;

    printk ( KERN_DEBUG "nanofs-readdir: Reading directory entry at blk_no %lxH\n",i->i_ino );


    //printk(KERN_DEBUG "nanofs: Adding files\n");
    //if(filldir(dirent, ".", 1, offset, de->d_inode->i_ino, DT_DIR)<0)
    //  return stored;
    //if(filldir(dirent, "..", 2, offset, de->d_parent->d_inode->i_ino, //DT_DIR)<0)
    //    return stored;
    ino=i->i_ino;
    if ( nanofs_fill_dir_node ( i->i_sb,ino,&dn ) ) { // parent dir
        return 0;
    }

    ino=dn.d_data_ptr;
    if ( nanofs_fill_dir_node ( i->i_sb,ino,&dn ) ) { // first filename
        return 0;
    }

    printk ( KERN_DEBUG "nanofs-readdir: Directory entry, len=%d\n",dn.d_fname_len );
    if ( DN_ISREG(dn) )  { // Regular file
        if ( filldir ( dirent, dn.d_fname, dn.d_fname_len, filp->f_pos, ino,DT_REG ) ) // over
            return 0;

    } else {
        if ( filldir ( dirent, dn.d_fname, dn.d_fname_len, filp->f_pos, ino,DT_DIR ) ) // over
            return 0;
    }


    //return 0; // 0 => End of directory reached
    // return the next directory entry of the dir directory
    // Next entry at:
    //if(dir_node.next_p==0)
    //{
    //  printk(KERN_DEBUG "nanofs: readdir end");
    //  return 0;
    //}
    filp->f_pos++;
    return 1;  // Success
}


/* Get inode_ino eq to blk_no on nanofs
 */
static struct inode *nanofs_get_inode ( struct super_block *sb, unsigned long pos ) {
    int res=0;
    struct nanofs_dir_node dn;
    struct inode *i;

    // read dir_node from blockdevice
    res=nanofs_fill_dir_node ( sb,pos,&dn );

    if ( res<0 )
        return ERR_PTR ( -EIO );

    i=iget_locked ( sb, pos );
    if ( !i )
        return ERR_PTR ( -ENOMEM );

    //if (!(i->i_state & I_NEW))
    //  return i;

    i->i_nlink = 1;
    i->i_ino=pos;


    i->i_mtime.tv_sec = i->i_atime.tv_sec = i->i_ctime.tv_sec = 0;
    i->i_mtime.tv_nsec = i->i_atime.tv_nsec = i->i_ctime.tv_nsec = 0;

    if ( DN_ISREG(dn)) { // 1: reg file
        printk ( KERN_DEBUG "nanofs-get-inode: found file entry at inode/blk_no %lxH\n",pos );
        i->i_mode = S_IFREG | S_IRUSR | S_IWUSR;
        i->i_fop  = &nanofs_file_ops;
        i->i_data.a_ops = &nanofs_aops;
        i->i_size = nanofs_get_file_size(sb,&dn); // FIXME: Read nanofs_sb to get it
        i->i_blocks = 1;
    } else { // Directory
        printk ( KERN_DEBUG "nanofs-get-inode: found dir entry at inode/blk_no %lxH\n",pos );
        i->i_mode = S_IFDIR| S_IRUSR | S_IWUSR;
        i->i_op = &nanofs_dir_inode_ops;
        i->i_fop = &nanofs_dir_ops;
        i->i_size = NANOFS_BLKSIZE; // FIXME: Read nanofs_sb to get it
    }


    unlock_new_inode ( i );

    return i;
}

static int nanofs_fill_sb ( struct super_block *sb, void *data, int silent )
{
    struct buffer_head *bh;
    struct inode *nanofs_root_inode=NULL;
    struct nanofs_superblock *nanofs_sb=NULL;
    //__u8 *bh_raw=NULL;
    int error=0;

    // Memory for nanofs info struct
    nanofs_sb = kzalloc ( sizeof ( struct nanofs_superblock ), GFP_KERNEL );

    if ( !nanofs_sb )
        return -ENOMEM;

    if ( sb == NULL ) {
        printk ( KERN_DEBUG "nanofs-fill-sb: Error superblock is NULL\n" );
        return -EINVAL;
    }


    // Superblock config
    sb->s_maxbytes = 0xFFFFFFFF;
    sb->s_fs_info=nanofs_sb;
    sb->s_op = &nanofs_ops; // Superblock operations
    sb->s_magic = NANOFS_MAGIC;  // // Magic number
    sb->s_flags |= MS_RDONLY | MS_NOATIME; //FIXME:
    // sb->s_type=&nanofs_type; TODO: ??

    // Read block device
    sb_min_blocksize ( sb, NANOFS_BLKSIZE ); // Try set bs=512
    printk ( KERN_DEBUG "nanofs-fill-sb: superblock b_size=%ld",sb->s_blocksize );
    bh=sb_bread ( sb,0 );
    if ( !bh ) {
        printk ( KERN_DEBUG "nanofs-fill-sb: Error cannot read blk device" );
        kfree ( nanofs_sb );
        return -EIO;
    }

    if ( bh->b_size!=sb->s_blocksize ) {
        printk ( KERN_DEBUG "nanofs-fill-sb: Error dev_b_size!=cache_b_size" );
        error=-EIO;
        goto fail;

    }

    // Fill nano fs super block info from device
    memcpy(nanofs_sb,bh->b_data,sizeof(struct nanofs_superblock));

    printk ( KERN_DEBUG "nanofs-fill-sb: superblock info blksz=%xH, root_dir=%xH, freeblock=%xH",
             nanofs_sb->s_blocksize,nanofs_sb->s_alloc_ptr,nanofs_sb->s_free_ptr );

    printk ( KERN_NOTICE "nanofs-fill-sb: filesystem revision %x detected with %d blocks\n",
             nanofs_sb->s_revision,nanofs_sb->s_fs_size );

    // Some checks
    if (le16_to_cpu(nanofs_sb->s_magic)!=NANOFS_MAGIC)
    {
        printk (KERN_ERR "NANOFS: Not valid nanofilesystem on device, bad magic number (%x!=%x).\n",
                nanofs_sb->s_magic,NANOFS_MAGIC);
        return -EIO; // FIXME;
        goto fail;
    }
    if (nanofs_sb->s_revision>NANOFS_REVISION)
    {
        printk (KERN_ERR "NANOFS: Not supported nanofilesystem revision number.\n");
        goto fail;
    }
    if (nanofs_sb->s_blocksize!=1)
    {
        printk (KERN_ERR "NANOFS: Only block size of 512 bytes is supported.\n");
        goto fail;
    }
    else
    {
        sb->s_blocksize=NANOFS_BLKSIZE;
        sb->s_blocksize_bits = blksize_bits(NANOFS_BLKSIZE);
    }

    // Get root directory

    nanofs_root_inode=nanofs_get_inode ( sb, le32_to_cpu(nanofs_sb->s_alloc_ptr ) );

    sb->s_root= d_alloc_root ( nanofs_root_inode );

    if ( !sb->s_root ) {
        printk ( KERN_DEBUG "nanofs-fill-sb: Error getting root inode\n" );
        //return -ENOMEM;
        goto fail;
    }


    //sb->s_root = d_alloc_root(nanofs_root_inode);

    //if(!(sb->s_root)) {
    //      iput(nanofs_root_inode);
    //



    // Success
    //dump_bh(bh);
    brelse ( bh );
    printk ( KERN_DEBUG "nanofs-fill-sb: End with res=%d",error );
    return error;

fail:
    brelse ( bh );
    kfree ( nanofs_sb );
    return -EIO;

}

/*static void nanofs_put_super ( struct super_block *sb )
{

    //struct nanofs_super_block *nanofs_sb;

    //nanofs_sb=sb->s_fs_info;

    //kill_block_super(sb);

    //iput(sb->s_root);


    //sb->s_fs_info = NULL;

}*/

/*
 * destroy a romfs superblock in the appropriate manner
 */
static void nanofs_kill_sb ( struct super_block *sb )
{
    struct nanofs_super_block *nanofs_sb;
    if ( sb->s_fs_info!=0 ) {
        nanofs_sb=sb->s_fs_info;
        kfree ( nanofs_sb );
    }
    kill_block_super ( sb );

}


static struct dentry *nanofs_mount ( struct file_system_type *fs_type,
                                                 int flags, const char *dev_name,
                                                 void *data ) {
    struct dentry *ret = ERR_PTR ( -EINVAL );
    // Mount on block
    ret = mount_bdev ( fs_type, flags, dev_name, data, nanofs_fill_sb );
    return ret;

}



/** Init Module */

static struct file_system_type nanofs_type = {
    .owner  = THIS_MODULE,
    .name   = "nanofs",
    .mount  = nanofs_mount,
    .kill_sb    = nanofs_kill_sb,
    .fs_flags   = FS_REQUIRES_DEV,
};



int __init nanofs_init ( void )
{
    printk ( KERN_DEBUG "nanofs-init: Debug printk enabled\n" );
    printk ( KERN_NOTICE "nanofs-init: registering filesystem\n" );
    return register_filesystem ( &nanofs_type );
}

static void __exit nanofs_exit ( void )
{
    unregister_filesystem ( &nanofs_type );
    printk ( KERN_NOTICE "nanofs-exit: nanofs unloaded\n" );
}

module_init ( nanofs_init );
module_exit ( nanofs_exit );

/** End init module */


