/*****************************************************************************
 * Author: Paulino Ruiz de Clavijo Vázquez <paulino@dte.us.es>
 * Date: 05-12-2012
 * Revision: 0.3
 * Project: Nanofs
 /** -- End auto header, don't touch this line -- ****************************/

#define _LARGEFILE64_SOURCE
#include<asm/types.h>
#include <sys/types.h>
#include<string.h>
#include<malloc.h>
#include<fcntl.h>
#include<stdio.h>

#include <unistd.h>

#include"nanofs_filedir.h"

#define NANOFSMGR_IO_BUFFER 1024

static int nanofsmgr_get_file_size(struct nanofs_file_info *fi_io, struct nanofs_fs_handle *hd);
static int nanofsmgr_alloc_data_node(struct nanofs_file_info *fi, struct nanofs_data_node *dn_io, struct nanofs_fs_handle *hd);
static int nanofsmgr_alloc_dir_node(struct nanofs_dir_node *dir_node_in, struct nanofs_fs_handle *hd);

//static int nanofsmgr_free_dir_node(int blkno, struct nanofs_fs_handle *hd);


/** Open device/file and fill handle */
int nanofs_open_dev(char *dev_name, struct nanofs_fs_handle *hd)
{
    //hd->h_fd=open ( dev_name, O_FSYNC  | O_RDWR, 0 );
    hd->h_fd = open(dev_name, O_RDWR, 0);
    hd->h_error = 0;
    if (hd->h_fd < 0)
    {
        fprintf(stderr, "** Cannot open device for R/W\n");
        hd->h_error = NANOFSMGR_ERR_IO;
    }
    if (hd->h_error == 0) // Ok
    {
        hd->h_dev_name = strdup(dev_name);
        // Reading superblock
        if (nanofs_read_sb(hd->h_fd, (off64_t) 0, &hd->h_sb) != 0)
        {
            fprintf(stderr, "** Error reading superblock\n");
            hd->h_error = NANOFSMGR_ERR_IO;
        }
    }

    // Check superblock
    if (hd->h_error == 0 && (hd->h_sb).s_magic != NANOFS_MAGIC)
    {
        fprintf(stderr, "** Error in superblock, bad magic number\n");
        hd->h_error = NANOFSMGR_ERR_SB;
    }
    // Get block bits
    if (hd->h_error == 0)
    {
        hd->h_block_bits = nanofs_get_block_bits(&(hd->h_sb));
        if (hd->h_block_bits < 0)
        {
            fprintf(stderr, "** Error in superblock getting blockbits\n");
            hd->h_error = NANOFSMGR_ERR_SB;
        }
    }
    if (hd->h_error == 0)
        nanofs_chdir("/", hd);

    if (hd->h_error != 0)
    {
        nanofs_close_dev(hd);
        return -1;
    }
    return 0;
}

int nanofs_close_dev(struct nanofs_fs_handle *hd)
{
    if (hd->h_fd < 0)
        return -1;
    close(hd->h_fd);
    free(hd->h_dev_name);
    hd->h_fd = -1;
    return 0;

}

/** Returns free space avaliable on filesystem
 * @return Free space in bytes or -1 on fail
 */

int nanofs_free(struct nanofs_fs_handle *hd)
{
    int free_size = 0;
    __u32 blk_no = hd->h_sb.s_free_ptr;
    struct nanofs_data_node data_node;
    while (blk_no > 0)
    {
        if (nanofs_read_data_block(hd->h_fd, (off64_t)(blk_no) << hd->h_block_bits, &data_node) < 0)
        {
            fprintf(stderr, "** [Err:%d] Error reading free space structure\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_FREE;
            return -1;
        }
        free_size += data_node.d_len;
        blk_no = data_node.d_next_ptr;
    }
    return free_size;
}

/** Change to selected path
 * @return 0 on success, -1 on fail
 */

int nanofs_chpath(char *full_path, struct nanofs_fs_handle *hd)
{
    int err = 0;
    char *pos;
    char buf[NANOFS_MAXFILENAME + 1];
    int i;
    if (full_path[0] != '/')
    {
        fprintf(stderr, "** Error: dir name must be absolute path\n");
        return -1;
    }
    pos = full_path + 1;
    buf[0] = 0;
    for (i = 0;; i++)
    {
        if ((*pos) == '/')
        {
            if (nanofs_chdir(buf, hd) != 0)
            {
                fprintf(stderr, "** Error in chdir '%s'\n", buf);
                return -1;
            }
            i = 0;
            buf[0] = 0;
            pos++;
        }
        if ((*pos) == 0)
            return nanofs_chdir(buf, hd);

        buf[i] = *pos;
        buf[i + 1] = 0;
        pos++;
    }
    return 0;
}

int nanofs_chdir(char *dir_name, struct nanofs_fs_handle *hd)
{
    int blk_no = hd->h_sb.s_alloc_ptr;
    struct nanofs_dir_node dir_n;
    struct nanofs_file_info f_info;
    if (hd->h_error == 0 && strcmp(dir_name, "/") == 0) // go to root dir
    {
        if (nanofs_read_dir_node(hd->h_fd, (off64_t)(blk_no) << hd->h_block_bits, &dir_n) < 0)
        {
            fprintf(stderr, "** Error reading root dir node\n");
            hd->h_error = NANOFSMGR_ERR_SB;
        }
        else
        {
            if (DN_ISDIR(dir_n))
                hd->h_blk_no = blk_no; // Set cursor
            else
            {
                hd->h_error = NANOFSMGR_ERR_SB;
            }
        }
    }
    else if (nanofs_lookup(dir_name, &f_info, hd) == 0)
    {
        if (DN_ISDIR(f_info.f_dir_node))
        {
            hd->h_blk_no = f_info.f_blk_no;
        }
        else
        {
            fprintf(stderr, "** Error: '%s' is not a directory\n", dir_name);
            return -1;
        }
    }
    if (hd->h_error != 0)
        return -1;
    return 0;
}

int nanofs_mkdir(char *dir_name, struct nanofs_fs_handle *hd)
{
    int err = 0;
    int new_blkno;
    struct nanofs_dir_node dir_n;
    struct nanofs_file_info fi;

    // Check if directory already exists
    if (nanofs_lookup(dir_name, &fi, hd) == 0) // Found
    {
        fprintf(stderr, "** Error: Directory already exists\n");
        return -1;
    }
    if (hd->h_error != 0)
        return -1;
    dir_n.d_flags = 1; // Is directory
    dir_n.d_data_ptr = 0;
    dir_n.d_meta_ptr = 0;
    dir_n.d_next_ptr = 0;
    //TODO: Check if dir_name contains '/'
    if (strlen(dir_name) > NANOFS_MAXFILENAME)
    {
        fprintf(stderr, "** Warning dir name has been truncated, max filename length reached\n");
        dir_n.d_fname_len = 255;
    }
    else
        dir_n.d_fname_len = strlen(dir_name);
    strncpy(dir_n.d_fname, dir_name, NANOFS_MAXFILENAME);
    new_blkno = nanofsmgr_alloc_dir_node(&dir_n, hd);

    if (hd->h_error != 0)
        return -1;
    return 0;
}

/** Remove dir
 *  @param dir_name dir name
 *  @return 0 on success, -1 on fail
 * */
int nanofs_rmdir(char *dir_name, struct nanofs_fs_handle *hd)
{
    int blkno, prev_blkno, next_blkno;
    int block_bits = hd->h_block_bits;
    struct nanofs_dir_node dir_nd, prev_dir_nd;
    struct nanofs_file_info fi;

    if(nanofs_lookup(dir_name,&fi, hd)!=0)
    {
        fprintf(stderr, "** Error: directory not exist\n", __LINE__);
        return -1;
    }
    return nanofsmgr_free_dir_node(&fi,hd);


    /*
     OLD implementation
     // Read parent dir node
    if (nanofs_read_dir_node(hd->h_fd, (off64_t)(hd->h_blk_no) << block_bits, &dir_nd) != 0)
    {
        fprintf(stderr, "** [Err:%d] Error in 'rmdir' while reading directory node\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_DIR;
        return -1;
    }
    // Reading childs dir nodes
    prev_blkno = 0;
    blkno = dir_nd.d_data_ptr; // First child dir_entry
    memcpy(&prev_dir_nd, &dir_nd, sizeof(struct nanofs_dir_node));
    while (blkno != 0)
    {
        if (nanofs_read_dir_node(hd->h_fd, (off64_t)(blkno) << block_bits, &dir_nd) != 0)
        {
            fprintf(stderr, "** [Err:%d] Error in 'rmdir' while reading directory node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_DIR;
            return -1;
        }
        if (strcmp(dir_name, dir_nd.d_fname) == 0) // Found
        {
            break;
        }
        prev_blkno = blkno;
        blkno = dir_nd.d_next_ptr;
        memcpy(&prev_dir_nd, &dir_nd, sizeof(struct nanofs_dir_node));
    }
    if (blkno == 0)
    {
        fprintf(stderr, "** Error: directory not exist\n", __LINE__);
        return -1;

    }
    else // Found
    {
        if (nanofsmgr_free_dir_node(blkno, hd) < 0)
            return -1;
        if (prev_blkno == 0) // There is the first entry in dir_entry list
        {
            // Updating parent dir
            prev_dir_nd.d_data_ptr = dir_nd.d_next_ptr;
            if (nanofs_write_dir_node(hd->h_fd, (off64_t)(hd->h_blk_no) << block_bits, &prev_dir_nd) < 0)
            {
                fprintf(stderr, "** [Err:%d] Error in 'rmdir' while updating directory node\n", __LINE__);
                hd->h_error = NANOFSMGR_ERR_DIR;
                return -1;
            }
        }
        else
        {
            prev_dir_nd.d_next_ptr = dir_nd.d_next_ptr;
            if (nanofs_write_dir_node(hd->h_fd, (off64_t)(prev_blkno) << block_bits, &prev_dir_nd) < 0)
            {
                fprintf(stderr, "** [Err:%d] Error in 'rmdir' while updating directory node\n", __LINE__);
                hd->h_error = NANOFSMGR_ERR_DIR;
                return -1;
            }
        }

    }
    return 0;*/
}

/** Remove file
 * @param file_name file name on current directory
 */
int nanofs_rm(char *file_name,struct nanofs_fs_handle *hd)
{
    struct nanofs_file_info fi;
    struct nanofs_data_node data_nd;
    __u32 data_blkno;


    if(nanofs_lookup(file_name, &fi, hd) != 0)
        return -1;
    data_blkno=fi.f_dir_node.d_data_ptr;
    while(data_blkno!=0) // File has data blocks
    {
        if (nanofs_read_data_block(hd->h_fd, (off64_t)(data_blkno) << hd->h_block_bits, &data_nd) != 0)
        {
           fprintf(stderr, "** [Err:%d] IO error freeing data file\n", __LINE__);
           hd->h_error = NANOFSMGR_ERR_IO;
           return -1;
        }
        if(nanofsmgr_free_data_node(data_blkno,&data_nd,hd)!=0)
            return -1;
        data_blkno=data_nd.d_next_ptr;  // Next block
    }
    // Free dir entry
    if(nanofsmgr_free_dir_node(&fi,hd)!=0)
        return -1;

    return 0;
}


/** List current dir filling fi_vec
 * @return number of items or <0 on error: -(number of errors found) */
int nanofs_list_dir(struct nanofs_file_info fi_vec[], int vec_size, struct nanofs_fs_handle *hd)
{
    struct nanofs_dir_node dir_n;
    int items = 0;
    int block_bits = hd->h_block_bits;
    int blk_no;
    int errs = 0;
    // Read parent dir node
    if (nanofs_read_dir_node(hd->h_fd, (off64_t)(hd->h_blk_no) << block_bits, &dir_n) != 0)
    {
        fprintf(stderr, "** [Err:%d] Error in 'list_dir' while reading directory node\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_DIR;
        return -1;
    }
    // Reading childs dir nodes
    blk_no = dir_n.d_data_ptr;
    while (blk_no != 0 && items < vec_size)
    {
        if (nanofs_read_dir_node(hd->h_fd, (off64_t)(blk_no) << block_bits, &dir_n) != 0)
        {
            fprintf(stderr, "** [Err:%d] Error in 'list_dir' while reading directory node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_DIR;
            return -1;
        }
        memcpy(&(fi_vec[items].f_dir_node), &dir_n, sizeof(struct nanofs_dir_node));
        fi_vec[items].f_blk_no = blk_no;
        //fi_vec[items].f_size=0;
        errs -= nanofsmgr_get_file_size(&fi_vec[items], hd);
        items++;
        blk_no = dir_n.d_next_ptr;
    }
    if (errs < 0)
        return errs;
    return items;
}

/** Lookup dir entry
 * @return -1 on fail or 0 on success
 */

int nanofs_lookup(char *file_name, struct nanofs_file_info *fi, struct nanofs_fs_handle *hd)
{
    struct nanofs_dir_node dir_n;
    int block_bits = hd->h_block_bits;
    int blk_no;

    // Read parent dir node
    fi->f_parent_dir_blk_no=hd->h_blk_no;
    if (nanofs_read_dir_node(hd->h_fd, (off64_t)(hd->h_blk_no) << block_bits, &dir_n) != 0)
    {
        fprintf(stderr, "** [Err:%d] Error in 'list_dir' while reading directory node\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_DIR;
        return -1;
    }
    // Reading childs dir nodes
    blk_no = dir_n.d_data_ptr;
    fi->f_prev_dir_blk_no=0;
    while (blk_no != 0)
    {
        if (nanofs_read_dir_node(hd->h_fd, (off64_t)(blk_no) << block_bits, &dir_n) != 0)
        {
            fprintf(stderr, "** [Err:%d] Error in 'list_dir' while reading directory node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_DIR;
            return -1;
        }
        if (strcmp(file_name, dir_n.d_fname) == 0) // Found
        {
            fi->f_blk_no = blk_no;
            fi->f_size = 0;
            memcpy(&(fi->f_dir_node), &dir_n, sizeof(struct nanofs_dir_node));
            return 0;
        }
        fi->f_prev_dir_blk_no=blk_no;
        blk_no = dir_n.d_next_ptr;
    }
    return -1;

}

/** Copy file in current directory
 * @param local_file_name full path of local file to copy
 * @param dst_file_name destination file name, not full path
 * @return 0 on success, -1 on fail
 */
int nanofs_put_file(char* local_file_name, char* dst_file_name, struct nanofs_fs_handle* hd)
{
    int fd_file;
    int err = 0;
    struct nanofs_superblock sb;
    struct nanofs_data_node free_nd, data_nd;
    struct nanofs_file_info fi;
    int bytes = 0;

    off64_t file_size, offset;
    char *buf;

    // Some checks
    if (strchr(dst_file_name, '/') != NULL)
    {
        fprintf(stderr, "** Error: Character '/' not allowed for destination file name\n");
        return -1;
    }

    fd_file = open(local_file_name, O_RDONLY);
    if (fd_file < 0)
    {
        fprintf(stderr, "** Error: cannot open local file '%s'\n", local_file_name);
        return -1;
    }

    // Check if file exists
    if (nanofs_lookup(dst_file_name, &fi, hd) == 0) // Found
    {
        fprintf(stderr, "** Error: File already exists.\n");
        err = -1;
    }

    if (err == 0)
    {
        fi.f_dir_node.d_flags = 0; // REG FILE
        fi.f_dir_node.d_data_ptr = 0;
        fi.f_dir_node.d_meta_ptr = 0;
        fi.f_dir_node.d_next_ptr = 0;
        fi.f_dir_node.d_fname_len = strlen(dst_file_name);
        if (strlen(dst_file_name) > NANOFS_MAXFILENAME)
        {
            fprintf(stderr, "** Warning filename has been truncated, max filename length reached\n");
            fi.f_dir_node.d_fname_len = NANOFS_MAXFILENAME;
        }
        // Filename/basename
        stpncpy(fi.f_dir_node.d_fname, dst_file_name, NANOFS_MAXFILENAME);

        // Allocating and writting new dir entry at root dir
        fi.f_blk_no = nanofsmgr_alloc_dir_node(&fi.f_dir_node, hd);
        if (fi.f_blk_no == 0)
        {
            fprintf(stderr, "** Error: Failed allocating new directory entry\n");
            err = -1;
        }
        printf_verbose(1, "Allocated new dir entry at block %xH\n", fi.f_blk_no);
    }
    if (err == 0)
    {
        // Get local file size
        file_size = lseek64(fd_file, (off64_t) 0, SEEK_END);
        if (file_size <= 0)
        {
            fprintf(stderr, "** Error: Cannot get local file size\n");
            err = -1;
        }
    }

    if (err == 0)
    {
        printf_verbose(1, "File size is %ld, allocating space in nanofs filesystem ... ", file_size);
        // Allocating new data block for dir entry
        data_nd.d_len = file_size;
        fi.f_dir_node.d_data_ptr = nanofsmgr_alloc_data_node(&fi, &data_nd, hd);
        if (fi.f_dir_node.d_data_ptr <= 0)
        {
            fprintf(stderr, "** Error: Cannot allocate free space in device/file\n");
            err = -1;
        }
        else if (data_nd.d_len != file_size)
        {
            fprintf(stderr, "** [Err %d] File fragmentation is not implemented\n", __LINE__);
            err = -1;
        }
        // Update dir_entry
        if (nanofs_write_dir_node(hd->h_fd, (off64_t)(fi.f_blk_no) << hd->h_block_bits, &fi.f_dir_node) != 0)
        {
            fprintf(stderr, "** IO Error updating dir node\n", __LINE__);
            err = -1;
        }

    }

    if (err == 0)
    {
        // Copy file
        buf = malloc(NANOFSMGR_IO_BUFFER);
        lseek64(fd_file, 0, SEEK_SET);
        lseek64(hd->h_fd, ((off64_t)(fi.f_dir_node.d_data_ptr) << hd->h_block_bits) + 8, SEEK_SET); // data bytes
        printf_verbose(1, "writing data ");
        for (offset = 0; offset < file_size;)
        {
            bytes = read(fd_file, buf, NANOFSMGR_IO_BUFFER);
            printf_verbose(1, ".");
            fflush(stdout);
            if (bytes == 0)
                break;
            else if (bytes < 0 || write(hd->h_fd, buf, bytes) < 0)
            {
                fprintf(stderr, "** IO error in device/file\n");
                err = -1;
            }
            offset += bytes;
        }
        printf_verbose(1, " writes end\n");
        free(buf);
    }

    close(fd_file);
    return err;
}

/** Copy file from nanofs to local file
 * @param local_file_name file where the data will be written, destination file is overwritten
 * @param file_name filename in current dir
 * @return 0 on success, -1 on fail
 */
int nanofs_get_file(char* local_file_name, char* file_name, struct nanofs_fs_handle* hd)
{
    int file_fd, err = 0;
    struct nanofs_file_info fi;
    struct nanofs_data_node data_nd;
    size_t bytes, bytes_left;
    int data_blkno;
    char *io_buf;
    if (nanofs_lookup(file_name, &fi, hd) != 0)
    {
        fprintf(stderr, "** Error: file '%s' not exists\n", file_name);
        return -1;
    }
    //file_fd=open(local_file_name,O_CREAT|O_WRONLY,S_IRUSR|S_IWUSR);
    file_fd = creat(local_file_name, S_IRUSR | S_IWUSR);
    if (file_fd < 0)
    {
        fprintf(stderr, "** Error: cannot write in local file '%s'\n", local_file_name);
        return -1;
    }
    io_buf = malloc(NANOFSMGR_IO_BUFFER);
    data_blkno = fi.f_dir_node.d_data_ptr;
    while (data_blkno != 0)
    {
        if (nanofs_read_data_block(hd->h_fd, (off64_t)(data_blkno) << hd->h_block_bits, &data_nd) != 0)
        {
            fprintf(stderr, "** [Err:%d] IO error getting file\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_IO;
            return -1;
        }
        lseek64(hd->h_fd, ((off64_t)(data_blkno) << hd->h_block_bits) + 8, SEEK_SET); // data bytes
        bytes_left = data_nd.d_len;
        while (bytes_left > 0)
        {
            bytes = bytes_left > NANOFSMGR_IO_BUFFER ? NANOFSMGR_IO_BUFFER : bytes_left;
            bytes = (size_t) read(hd->h_fd, io_buf, bytes);
            if (write(file_fd, io_buf, bytes) != bytes)
            {
                fprintf(stderr, "** Error writing data to file\n");
                err = -1;
                break;
            }
            bytes_left -= bytes;
        }
        data_blkno = data_nd.d_next_ptr;
    }
    close(file_fd);
    free(io_buf);
    return err;
}

/** Calc file size
 */
static int nanofsmgr_get_file_size(struct nanofs_file_info *fi_io, struct nanofs_fs_handle *hd)
{
    struct nanofs_data_node data_nd;
    if (DN_ISDIR(fi_io->f_dir_node))
    {
        fi_io->f_size = 1 << hd->h_block_bits;
        return 0;
    }

    if (fi_io->f_dir_node.d_data_ptr == 0)
    {
        fi_io->f_size = 0;
        return 0;
    }
    else if (nanofs_read_data_block(hd->h_fd, (off64_t)(fi_io->f_dir_node.d_data_ptr) << hd->h_block_bits, &data_nd) != 0)
    {
        fprintf(stderr, "** [Err:%d] IO error getting file size\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    fi_io->f_size = data_nd.d_len;
    while (data_nd.d_next_ptr != 0)
    {
        if (nanofs_read_data_block(hd->h_fd, (off64_t)(data_nd.d_next_ptr) << hd->h_block_bits, &data_nd) != 0)
        {
            fprintf(stderr, "** [Err:%d] IO error getting file size\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_IO;
            return -1;
        }
        fi_io->f_size += data_nd.d_len;
    }
    return 0;
}

/** Allocate new dn at the end of current dir and return block_no of the new allocated dir node
 *  The new dir entry is not writen
 *  @return block_no allocated or 0 on fail
 */

static int nanofsmgr_alloc_dir_node(struct nanofs_dir_node *dir_node_in, struct nanofs_fs_handle *hd)
{
    struct nanofs_dir_node parent_node, dir_n;
    struct nanofs_data_node data_n;
    int block_bits = hd->h_block_bits;
    int new_dn_blkno, current_dn;

    if (nanofs_read_dir_node(hd->h_fd, (off64_t)(hd->h_blk_no) << block_bits, &parent_node) != 0)
    {
        fprintf(stderr, "** [Err:%d] Error in 'alloc_dir_node' while reading directory node\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_DIR;
        return 0;
    }

    // TODO: check after read if DN is a directory entry

    // Read free list

    if (nanofs_read_data_block(hd->h_fd, (off64_t)(hd->h_sb.s_free_ptr) << block_bits, &data_n) != 0)
    {
        fprintf(stderr, "** [Err:%d] Error in 'alloc_dir_node' while getting free blocks\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return 0;
    }

    if (data_n.d_len >> block_bits <= 0) // len is in bytes
    {
        fprintf(stderr, "** Error not enough free space on device to alloc new directory entry\n");
        return 0;
    }

    new_dn_blkno = hd->h_sb.s_free_ptr;
    hd->h_sb.s_free_ptr = hd->h_sb.s_free_ptr + 1; // FIXME: Only allowed 512 bytes block
    // Reduce free list by one block
    data_n.d_len = data_n.d_len - (__u32 ) (1 << block_bits);
    printf_verbose(1, "Updating free blocks\n");
    if (nanofs_write_data_block(hd->h_fd, (off64_t)(hd->h_sb.s_free_ptr) << block_bits, &data_n) != 0)
    {
        fprintf(stderr, "** [Err:%d] Cannot update freeblocks list, filesystem may be corrupted\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return 0;
    }

    // Update superblock
    printf_verbose(1, "Updating superblock\n");
    if (nanofs_write_sb(hd->h_fd, (off64_t) 0, &hd->h_sb) != 0)
    {
        fprintf(stderr, "** [Err:%d] Cannot update superblock, filesystem may be corrupted\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return 0;
    }

    // Add new dir entry at the end of list
    if (parent_node.d_data_ptr == 0) // Empty child list
    {
        parent_node.d_data_ptr = new_dn_blkno;
        // Add child dir_node
        printf_verbose(1, "Updating parent dir entry\n");
        if (nanofs_write_dir_node(hd->h_fd, (off64_t)(hd->h_blk_no) << block_bits, &parent_node) != 0)
        {
            fprintf(stderr, "** [Err:%d] Cannot update parent dir node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_IO;
            return 0;
        }
    }
    else
    {
        // follow list
        current_dn = parent_node.d_data_ptr;
        if (nanofs_read_dir_node(hd->h_fd, (off64_t)(current_dn) << block_bits, &parent_node) != 0)
        {
            fprintf(stderr, "** [Err:%d] Error in 'alloc_dir_node' while reading directory node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_DIR;
            return 0;
        }
        while (parent_node.d_next_ptr != 0)
        {
            current_dn = parent_node.d_next_ptr;
            if (nanofs_read_dir_node(hd->h_fd, (off64_t)(current_dn) << block_bits, &parent_node) != 0)
            {
                fprintf(stderr, "** [Err:%d] Error in 'alloc_dir_node' while reading directory node\n", __LINE__);
                hd->h_error = NANOFSMGR_ERR_DIR;
                return 0;
            }
        }
        // Updating current dir entry
        printf_verbose(1, "Adding dir entry to dir list\n");
        parent_node.d_next_ptr = new_dn_blkno;
        if (nanofs_write_dir_node(hd->h_fd, (off64_t)(current_dn) << block_bits, &parent_node) != 0)
        {
            fprintf(stderr, "** [Err:%d] Cannot update parent dir node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_IO;
            return 0;
        }

    }
    // Add new dir_node
    printf_verbose(1, "Writing new dir entry\n");
    if (nanofs_write_dir_node(hd->h_fd, (off64_t)(new_dn_blkno) << block_bits, dir_node_in) != 0)
    {
        fprintf(stderr, "** [Err:%d] Cannot write new dir node\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return 0;

    }
    return new_dn_blkno;
}

/** Alloc new block from free space at the end file, not writes data on new block.
 * @param fi dir entry which data is allocated
 * @param dn_io Datablock as input: dn_io.d_size = size required. As output dn_io.d_size is filled with size assigned
 * @return -1 on fail, blk_no allocated on success also
 * @todo Now Only allocates one block of dn_io.size, list allocating is not implemented.
 * */
static int nanofsmgr_alloc_data_node(struct nanofs_file_info *fi, struct nanofs_data_node *dn_io, struct nanofs_fs_handle *hd)
{
    struct nanofs_data_node free_nd;
    unsigned int new_free_blkno = 0, blocks_count;

    // Locating free space
    if (nanofs_read_data_block(hd->h_fd, (off64_t)(hd->h_sb.s_free_ptr) << hd->h_block_bits, &free_nd) != 0)
        return -1;

    if (dn_io->d_len > free_nd.d_len) // Fragmentation not supported
    {
        fprintf(stderr, "** [Err:%d] Fragmentation not supported\n", __LINE__);
        return -1;
    }
    // New alloc node is old free block
    if (fi->f_dir_node.d_data_ptr != 0)
    {
        fprintf(stderr, "** [Err:%d] File append not supported\n", __LINE__);
        return -1;
    }
    // Update file entry
    fi->f_dir_node.d_data_ptr = hd->h_sb.s_free_ptr;
    if (nanofs_write_dir_node(hd->h_fd, (off64_t)(fi->f_blk_no) << hd->h_block_bits, &(fi->f_dir_node)) != 0)
    {
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    // Writing new data block header
    dn_io->d_next_ptr = 0;
    if (nanofs_write_data_block(hd->h_fd, (off64_t)(fi->f_dir_node.d_data_ptr) << hd->h_block_bits, dn_io) != 0)
    {
        fprintf(stderr, "[Err:%d] IO Error writing blocks\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    // calc where new free block starts
    blocks_count = 1 + ((dn_io->d_len) >> hd->h_block_bits);
    printf_verbose(1, "allocated %d data blocks\n", blocks_count);
    hd->h_sb.s_free_ptr = hd->h_sb.s_free_ptr + blocks_count;
    free_nd.d_len = free_nd.d_len - (blocks_count << hd->h_block_bits);
    if (nanofs_write_data_block(hd->h_fd, (off64_t)(hd->h_sb.s_free_ptr) << hd->h_block_bits, &free_nd) != 0)
    {
        fprintf(stderr, "[Err:%d] IO Error writing free blocks\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    // Updating superblock
    if (nanofs_write_sb(hd->h_fd, (off64_t) 0, &(hd->h_sb)) != 0)
    {
        fprintf(stderr, "[Err:%d] IO Error updating superblock\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    return fi->f_dir_node.d_data_ptr;
}

/** Free dir node,
 * Dir node block is added at ahead of list of free nodes, linked list of parent dir is updated
 * @return 0 on success, -1 on fail
 * */
int nanofsmgr_free_dir_node(struct nanofs_file_info *fi, struct nanofs_fs_handle *hd)
{
    struct nanofs_data_node data_nd;
    struct nanofs_dir_node dir_nd;

    // First add this node to free nodes list
    data_nd.d_next_ptr = hd->h_sb.s_free_ptr;
    data_nd.d_len = (1 << hd->h_block_bits) - NANOFS_HDATA_NODE_SIZE;
    if (nanofs_write_data_block(hd->h_fd, (off64_t)(fi->f_blk_no) << hd->h_block_bits, &data_nd) < 0)
    {
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    hd->h_sb.s_free_ptr = fi->f_blk_no;
    if (nanofs_write_sb(hd->h_fd, (off64_t) 0, &hd->h_sb) != 0)
    {
        fprintf(stderr, "** [Err:%d] Cannot update superblock, filesystem may be corrupted\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    // Second update linked list

    if (fi->f_prev_dir_blk_no != 0) // There is the first entry in dir_entry list
    {
        // Updating prev dir node
        if (nanofs_read_dir_node(hd->h_fd, (off64_t)(fi->f_prev_dir_blk_no) << hd->h_block_bits, &dir_nd) != 0)
        {
           fprintf(stderr, "** [Err:%d] Error while freeing directory node\n", __LINE__);
           hd->h_error = NANOFSMGR_ERR_DIR;
           return -1;
        }
        dir_nd.d_next_ptr = fi->f_dir_node.d_next_ptr;
        if (nanofs_write_dir_node(hd->h_fd, (off64_t)(fi->f_prev_dir_blk_no) << hd->h_block_bits, &dir_nd) < 0)
        {
            fprintf(stderr, "** [Err:%d] Error in 'rmdir' while updating directory node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_DIR;
            return -1;
        }
    }
    else if (fi->f_parent_dir_blk_no!=0) // parent is not superblock
    {   // No previous entry, parent dir has no children
        // Read parent dir node
        if (nanofs_read_dir_node(hd->h_fd, (off64_t)(fi->f_parent_dir_blk_no) << hd->h_block_bits, &dir_nd) < 0)
        {
            fprintf(stderr, "** [Err:%d] Error in 'rmdir' while updating directory node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_DIR;
            return -1;
        }
        dir_nd.d_data_ptr=fi->f_dir_node.d_next_ptr;
        if (nanofs_write_dir_node(hd->h_fd, (off64_t)(fi->f_parent_dir_blk_no) << hd->h_block_bits, &dir_nd) < 0)
        {
            fprintf(stderr, "** [Err:%d] Error in 'rmdir' while updating directory node\n", __LINE__);
            hd->h_error = NANOFSMGR_ERR_DIR;
            return -1;
        }
    }
    else
    {
         fprintf(stderr, "** [Err:%d] Not implemented delete root dir entry\n", __LINE__);
         return -1;
    }

    return 0;
}

/** free data_node
 * data node is added at ahead of list of free nodes
 * @return 0 on success, -1 on fail
 * */

int nanofsmgr_free_data_node(int blkno,struct nanofs_data_node *dn, struct nanofs_fs_handle *hd)
{
    struct nanofs_data_node free_nd;

    // Expand len to block size
    free_nd.d_next_ptr = hd->h_sb.s_free_ptr;
    free_nd.d_len= ((((dn->d_len + NANOFS_HDATA_NODE_SIZE) >> hd->h_block_bits) +1) << hd->h_block_bits) - NANOFS_HDATA_NODE_SIZE;

    if (nanofs_write_data_block(hd->h_fd, (off64_t)(blkno) << hd->h_block_bits, &free_nd) < 0)
    {
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    hd->h_sb.s_free_ptr = blkno;
    if (nanofs_write_sb(hd->h_fd, (off64_t) 0, &hd->h_sb) != 0)
    {
        fprintf(stderr, "** [Err:%d] Cannot update superblock, filesystem may be corrupted\n", __LINE__);
        hd->h_error = NANOFSMGR_ERR_IO;
        return -1;
    }
    return 0;
}


