/*****************************************************************************
* Author: Paulino Ruiz de Clavijo Vázquez <paulino@dte.us.es>
* Date: 05-12-2012
* Revision: 0.3
* Project: Nanofs
/** -- End auto header, don't touch this line -- ****************************/
 
 


#define _LARGEFILE64_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm/types.h>
#include <stdio.h>
#include <libintl.h>
#include <locale.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libgen.h>

#include <nanofs.h>
#include <nanofs_filedir.h>
#include <nanofsmanager.h>
#include <utils.h>


/** Help str */
static const char *HelpStr = "Usage:"\
                             "\tnanofs-manager [OPTION] DEVICE get FILE DST_FILE\n"\
                             "\tnanofs-manager [OPTION] DEVICE put FILE DST_PATH\n"\
                             "\tnanofs-manager [OPTION] DEVICE rm  FILE_PATH\n"\
                             "\tnanofs-manager [OPTION] DEVICE mkdir DIR\n"\
                             "\tnanofs-manager [OPTION] DEVICE rmdir DIR\n"\
                             "\tnanofs-manager [OPTION] DEVICE ls DIR\n"\
                             "\tnanofs-manager [OPTION] DEVICE free DIR\n"\
                             "\nOptions\n"\
                             "\t-v verbose\n";

int global_verbose=0;


int main ( int argc, char **argv )
{
    int res=EXIT_SUCCESS;
    struct nanofs_fs_handle nfs_h;
    int opt_device=1;
    
    // Options
    for(opt_device=1;opt_device<argc;opt_device++)
    {
      if(argv[opt_device][0]!='-') // End of options
        break;
      
      if(argv[opt_device][1]=='v') 
        global_verbose=1;
      else
      {
        fprintf(stderr,"** Error: Unknown option %s at command line\n",argv[opt_device]);
        res=EXIT_FAILURE;
        break;
      }
    }
    if(res==EXIT_SUCCESS && argc<opt_device + 2)
    {
        printf(HelpStr,NULL);
        res=EXIT_FAILURE;
    }
    if(res==EXIT_SUCCESS)
    {
      nanofslib_set_verbosity(global_verbose);
      if(global_verbose)
        print_version(basename(argv[0]));
      if(nanofs_open_dev(argv[opt_device],&nfs_h)!=0)
      {
        fprintf(stderr,"** Error: Cannot open nanofs filesystem (%s)",argv[opt_device]);
        res=EXIT_FAILURE;
      }
    }
    if(res==EXIT_SUCCESS)
    {         
        if (strcmp(argv[opt_device+1],"put")==0)
            res=do_put(argv[opt_device+2],argv[opt_device+3],&nfs_h);
        else if (strcmp(argv[opt_device+1],"rm")==0)
            res=do_rm(argv[opt_device+2],&nfs_h);
        else if (strcmp(argv[opt_device+1],"mkdir")==0)
            res=do_mkdir(argv[opt_device+2],&nfs_h);
        else if (strcmp(argv[opt_device+1],"rmdir")==0)
            res=do_rmdir(argv[opt_device+2],&nfs_h);
        else if (strcmp(argv[opt_device+1],"ls")==0)
            res=do_ls(argv[opt_device+2],&nfs_h);
        else if (strcmp(argv[opt_device+1],"get")==0)
            res=do_get(argv[opt_device+2],argv[opt_device+3],&nfs_h);
        else if (strcmp(argv[opt_device+1],"free")==0)
            res=do_free(&nfs_h);	
        else
        {
            fprintf(stderr,"Command '%s' unknown\n",argv[opt_device+1]);
        }
        nanofs_close_dev(&nfs_h);
    }
    return res;
}

int do_free(struct nanofs_fs_handle *hd)
{
  int size=nanofs_free(hd);
  if(size>=0)
  {
    printf("%d Bytes\n",size);
    return 0;
  } 
  else
    return -1;
}

int do_ls(char *dir_name,struct nanofs_fs_handle *hd)
{
    int items=-1,i;
    struct nanofs_file_info f_v[1000];
    
    if(dir_name==NULL)
    {
        fprintf(stderr,"** Error: missing argument for ls command\n");
        return -1;
    }
    
    if(nanofs_chpath(dir_name,hd)==0)
      items=nanofs_list_dir(f_v,1000,hd);
    else
    {
        fprintf(stderr,"** Error: Cannot read dir '%s'\n",dir_name);
        return -1;
    }
    for(i=0;i<items;i++)
    {
        printf("%10ld  ",f_v[i].f_size);
        if(DN_ISREG(f_v[i].f_dir_node))
            printf("A ");
        else
            printf("D ");
        printf("%-s\n",f_v[i].f_dir_node.d_fname);
    }
    return 0;
}


int do_mkdir(char *dir_name,struct nanofs_fs_handle *hd)
{
    int err=0;
    char *pos;
    char buf[NANOFS_MAXFILENAME+1];
    int i;
    if(dir_name==NULL)
    {
        fprintf(stderr,"** Error: missing argument for 'mkdir' command\n");
        return -1;
    }
    if(dir_name[0]!='/')
    {
        fprintf(stderr,"** Error: dir name must be absolute path\n");
        return -1;
    }
    pos=dir_name +1;
    buf[0]=0;
    for(i=0;;i++)
    {
        if( (*pos)=='/')
        {
            if(nanofs_chdir(buf,hd)!=0)
            {
                fprintf(stderr,"** Error in chdir '%s'\n",buf);
                return -1;
            }
            i=0;
            buf[0]=0;
            pos++;
        }
        if( (*pos)==0)
            return nanofs_mkdir(buf,hd);
        
        buf[i]=*pos;
        buf[i+1]=0;
        pos++;
    }
    return 0;
}

/** rmdir 
 * @param file_path absolute directory path to remove
 * @param hs nanofs handle
 * */
int do_rmdir(char *dir_path,struct nanofs_fs_handle *hd)
{
    int err=0;
    char *pos;
    char buf[NANOFS_MAXFILENAME+1];
    char *dir_str,*base_str,*bname,*dname;

    int i;
    if(dir_path==NULL)
    {
        fprintf(stderr,"** Error: missing argument for 'rmdir' command\n");
        return -1;
    }
    if(dir_path[0]!='/')
    {
        fprintf(stderr,"** Error: dir name must be absolute path\n");
        return -1;
    }
    
    dir_str=strdup(dir_path);
    base_str=strdup(dir_path);
    bname=basename(base_str);
    dname=dirname(dir_str);
    if(nanofs_chpath(dname,hd)<0) // Chdir to parent directory
    {
        free(dir_str);
        free(base_str);
        fprintf(stderr,"** Error: path '%s' not exist\n",dname);
        return -1;        
    }
    if(nanofs_rmdir(bname,hd)<0) // Do rmdir
    {
        free(dir_str);
        free(base_str);
        fprintf(stderr,"** Error: cannot delete dir '%s'\n",bname);
        return -1;                
    }
    free(dir_str);
    free(base_str);
    return 0;
}
/** Delete file
 *
 */
int do_rm(char *file_path,struct nanofs_fs_handle *hd)
{
    int err=0;
    char *pos;
    char buf[NANOFS_MAXFILENAME+1];
    char *dir_str,*base_str,*bname,*dname;

    if(file_path==NULL)
    {
        fprintf(stderr,"** Error: missing argument for 'rm' command\n");
        return -1;
    }
    if(file_path[0]!='/')
    {
        fprintf(stderr,"** Error: file path must be absolute path\n");
        return -1;
    }


    dir_str=strdup(file_path);
    base_str=strdup(file_path);
    bname=basename(base_str);
    dname=dirname(dir_str);
    if(nanofs_chpath(dname,hd)<0) // Chdir to parent directory
    {
        free(dir_str);
        free(base_str);
        fprintf(stderr,"** Error: path '%s' not exist\n",dname);
        return -1;
    }
    if(nanofs_rm(bname,hd)<0) // Do rm
    {
        free(dir_str);
        free(base_str);
        fprintf(stderr,"** Error: cannot delete file '%s'\n",bname);
        return -1;
    }
    free(dir_str);
    free(base_str);
    return 0;
}

/** Put file to nanofs file system
 * @param file_name local file path
 * @param dts_path destination directory
 * */
int do_put(char *file_name,char* dst_path,struct nanofs_fs_handle *hd)
{
    char *buf;
    if(file_name==NULL || dst_path==NULL)
    {
        fprintf(stderr,"** Error: missing argument for 'put' command\n");
        return -1;
    }
    
    if(nanofs_chpath(dst_path,hd)!=0)
        return -1;
    
    buf=(char *)basename(file_name);
        
    return nanofs_put_file(file_name,buf,hd);
    

}

/** Put file to nanofs file system
 * @param file_name local file path
 * @param dts_path destination directory
 * */
int do_get(char *file_path,char* dst_path,struct nanofs_fs_handle *hd)
{
    char dir_name[NANOFS_MAX_PATH];
    char *file_name,*last=strrchr(file_path,'/');
    char *pos=file_path;
    int i;
    
    if(file_path==NULL || dst_path==NULL)
    {
        fprintf(stderr,"** Error: missing argument for 'get' command\n");
        return -1;
    }

    
    
    if(last!=NULL)
    {
        for(i=0;pos<=last && i < NANOFS_MAX_PATH - 1;i++)
        {
            dir_name[i]=*pos;
            dir_name[i+1]=0;
            pos++;
        }
        if(nanofs_chpath(dir_name,hd)!=0)
            return -1;
    }
    file_name=(char *)basename(file_path);
    if(file_name!=NULL)   
        return nanofs_get_file(dst_path,file_name,hd);
    return -1;
    

}

